﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Algorithms.EGO {
  [StorableClass]
  [Item("InfillProblem", "A problem for finding the most interesing potential new sampling Points by optimizing some InfillCriterion")]
  public sealed class InfillProblem : SingleObjectiveBasicProblem<RealVectorEncoding> {

    public override bool Maximization => true;  //This is necessary because algorithms do not expect the maximization to change

    #region Properties;
    [Storable]
    private IInfillCriterion infillCriterion;
    [Storable]
    private SingleObjectiveBasicProblem<IEncoding> problem;

    public IInfillCriterion InfillCriterion
    {
      get { return infillCriterion; }
      set { infillCriterion = value; }
    }
    public SingleObjectiveBasicProblem<IEncoding> Problem
    {
      get { return problem; }
      set
      {
        problem = value;
        if (problem == null) return;
        var enc = problem.Encoding as RealVectorEncoding;
        if (enc == null) throw new ArgumentException("EGO can not be performed on non-RealVectorEncodings");
        Encoding = enc;
        SolutionCreator = new UniformRandomRealVectorCreator();//ignore Problem specific Solution Creation

      }
    }
    #endregion

    #region HLConstructors
    [StorableConstructor]
    private InfillProblem(bool deserializing) : base(deserializing) { }
    private InfillProblem(InfillProblem original, Cloner cloner) : base(original, cloner) {
      infillCriterion = cloner.Clone(original.InfillCriterion);
      problem = cloner.Clone(original.Problem);
    }
    public InfillProblem() { }
    public override IDeepCloneable Clone(Cloner cloner) { return new InfillProblem(this, cloner); }
    #endregion

    public override double Evaluate(Individual individual, IRandom r) {
      var q = InfillCriterion.Evaluate(individual.RealVector());
      return InfillCriterion.Maximization() ? q : -q;
    }
    public override void Analyze(Individual[] individuals, double[] qualities, ResultCollection results, IRandom random) {
      base.Analyze(individuals, qualities, results, random);
      var best = qualities.ArgMax(x => x);
      var qnew = InfillCriterion.Maximization() ? qualities[best] : -qualities[best];
      const string qname = EfficientGlobalOptimizationAlgorithm.BestInfillQualityResultName;
      const string sname = EfficientGlobalOptimizationAlgorithm.BestInfillSolutionResultName;
      if (!results.ContainsKey(EfficientGlobalOptimizationAlgorithm.BestInfillQualityResultName)) {
        results.Add(new Result(sname, (RealVector)individuals[best].RealVector().Clone()));
        results.Add(new Result(qname, new DoubleValue(qnew)));
        return;
      }
      var qold = results[qname].Value as DoubleValue;
      if (qold == null) throw new ArgumentException("Old best quality is not a double value. Conflicting Analyzers?");
      if (qold.Value >= qnew == InfillCriterion.Maximization()) return;
      results[sname].Value = (RealVector)individuals[best].RealVector().Clone();
      qold.Value = qnew;
    }
  }
}
